"""
Pydantic classes for statistics which generate network graphs.
"""

from typing import List, Optional

from pydantic import BaseModel


class NetworkMetaData(BaseModel):
    """Generic class to hold the metadata for data/stats with a network like structure.
    this is used to generate api calls for sankey diagrams
    """

    title: Optional[str] = None
    total_population: Optional[int] = None


class Nodes(BaseModel):
    """Contains labels/names accosiated with each node in the network"""

    node_labels: Optional[List[str]] = None


class Vertices(BaseModel):
    """Contains data to create vertices connecting each of the nodes"""

    source: List[str]
    target: List[str]
    value: List[str]


class LabelledNetwork(BaseModel):
    metadata: NetworkMetaData
    node: Nodes
    link: Vertices



"""
Generic 3D/map data model for UKRDC stats
"""

from datetime import datetime
from typing import List, Union, Optional

from pydantic import BaseModel

Number = Union[int, float]


class TimeSeries3dData(BaseModel):
    x: List[datetime]
    y: List[Number]
    z: List[Number]


class AxisLabel3d(BaseModel):
    x: Optional[str]
    y: Optional[str]
    z: Optional[str]


class Basic3dMetadata(BaseModel):
    title: Optional[str] = None
    axis_titles: Optional[AxisLabel3d] = None


class TimeSeries3d(BaseModel):
    meta_data: Basic3dMetadata
    data: TimeSeries3dData



"""
Generates biomaker data for a list of patients.
The stats here are loosly based on the metrics of chapter 5 of the annual report.
"""

import datetime as dt
from typing import List, Union
import pandas as pd


from sqlalchemy import select, and_
from sqlalchemy.orm import Session
from ukrdc_sqla.ukrdc import ResultItem, LabOrder, CauseOfDeath, Patient, Code
from ukrdc_stats.exceptions import NoCohortError
from .models.generic_2d import Labelled2d, Labelled2dMetadata, Labelled2dData
from pydantic import BaseModel


# pydantic classes for the api
class BiomarkerMeta(BaseModel):
    title: str
    data_axis_titles: List[str]


class BiomarkerData(BaseModel):
    timestamps: List[dt.datetime]
    testresult: List[Union[float, None]]
    orderid: List[str]
    resultid: List[str]


class PatientBiomarker(BaseModel):
    ukrdcid: str
    median_result: Union[float, None]
    most_recent_result: Union[float, None]
    data: BiomarkerData


class PatientBiomarkers(BaseModel):
    meta_data: BiomarkerMeta
    patient_data: List[PatientBiomarker]


def produce_output(
    query_result: pd.DataFrame, meta_data: BiomarkerMeta
) -> PatientBiomarkers:

    # There is quite a bit of junk in the results table so this statement turns them
    # into nans
    query_result.resultvalue = query_result.resultvalue.apply(
        pd.to_numeric, errors="coerce"
    )

    # initialise variable for output
    biomarkers = PatientBiomarkers(meta_data=meta_data, patient_data=[])

    for ukrdcid in query_result.ukrdcid.unique():
        test_data_slice = (
            query_result[query_result.ukrdcid == ukrdcid]
            .replace("", None)
            .sort_values("enteredon")
        )

        biomarker = PatientBiomarker(
            ukrdcid=ukrdcid,
            median_result=test_data_slice.resultvalue.median(),
            most_recent_result=test_data_slice.resultvalue.iloc[-1],
            data=BiomarkerData(
                timestamps=test_data_slice.enteredon.tolist(),
                testresult=test_data_slice.resultvalue.tolist(),
                orderid=test_data_slice.orderid.tolist(),
                resultid=test_data_slice.resultid.tolist(),
            ),
        )

        biomarkers.patient_data.append(biomarker)

    return biomarkers


def generic_biomarkers_query(
    start_time: dt.datetime,
    end_time: dt.datetime,
    patients: pd.DataFrame,
    service_codes: List,
    session: Session,
) -> pd.DataFrame:
    """Queries the database to return a set of test results of a specified type for a given set of patients over a given timewidow.

    Args:
        start_time (dt.datetime): start of window
        end_time (dt.datetime): end of window
        patients (pd.DataFrame): list of patient identifiers must contain both the pid and the ukrdcid
        service_codes (List): list of service codes specifing the type of test
        session (Session):

    Returns:
        pd.DataFrame: result of queary
    """

    generic_query = (
        select(
            LabOrder.pid,
            ResultItem.id.label("resultid"),
            ResultItem.value,
            ResultItem.service_id,
            LabOrder.entered_on,
            ResultItem.order_id,
        )
        .join(LabOrder, LabOrder.id == ResultItem.order_id)
        .where(
            and_(
                LabOrder.pid.in_(patients.pid.tolist()),
                ResultItem.service_id.in_(service_codes),
                LabOrder.entered_on > start_time,
                LabOrder.entered_on < end_time,
            )
        )
    )

    test_data = pd.read_sql(generic_query, session.bind)
    test_data_merged = test_data.merge(patients, "inner", on="pid")

    return test_data_merged


def urea_reduction_ratio(
    start_time: dt.datetime,
    end_time: dt.datetime,
    patient: pd.DataFrame,
    session: Session,
) -> PatientBiomarkers:
    """_summary_

    Args:
        start_time (dt.datetime): Start date to collect tests from
        end_time (dt.datetime): Start date to collect tests from
        patient (pd.DataFrame): Dataframe containing pids and ukrdcids of patients this can be generated with the dialysis stats module
        session (Session): ukrdc3 session

    Returns:
        pd.DataFrame: Urea reduction ratio results for the above parameters
    """

    # query database to get test results
    test_data_id_merged = generic_biomarkers_query(
        start_time=start_time,
        end_time=end_time,
        patients=patient,
        service_codes=["QBLG9"],
        session=session,
    )

    if test_data_id_merged is None:
        raise NoCohortError("No URR results extracted")

    return produce_output(
        test_data_id_merged,
        BiomarkerMeta(
            title="Urea Reduction Ratio",
            data_axis_titles=["Date", "URR", "Order ID", "Test ID"],
        ),
    )


def haemoglobin(
    start_time: dt.datetime,
    end_time: dt.datetime,
    patient: pd.DataFrame,
    session: Session,
):

    # query haemoglobin
    test_data_id_merged = generic_biomarkers_query(
        start_time=start_time,
        end_time=end_time,
        patients=patient,
        service_codes=["QBLE1", "QBLEB"],
        session=session,
    )

    # ensure units match up
    test_data_id_merged.resultvalue.apply(pd.to_numeric, errors="coerce")
    test_data_id_merged[test_data_id_merged.serviceidcode == "QBLE1"].resultvalue = (
        test_data_id_merged[test_data_id_merged.serviceidcode == "QBLE1"].resultvalue
        * 10.0
    )

    return produce_output(
        test_data_id_merged,
        BiomarkerMeta(
            title="Haemoglobin",
            data_axis_titles=["Date", "Haemoglobin", "Order ID", "Test ID"],
        ),
    )


def cause_of_death(
    start_time: dt.datetime,
    end_time: dt.datetime,
    patient: pd.DataFrame,
    session: Session,
):

    # query cause of death
    query_cod = (
        select(
            Patient.death_time,
            CauseOfDeath.pid,
            CauseOfDeath.diagnosis_code,
            CauseOfDeath.diagnosis_code_std,
            Code.description.label("cause_of_death"),
        )
        .join(Patient, Patient.pid == CauseOfDeath.pid)
        .join(
            Code,
            and_(
                Code.coding_standard == CauseOfDeath.diagnosis_code_std,
                Code.code == CauseOfDeath.diagnosis_code,
            ),
        )
        .where(
            and_(
                CauseOfDeath.pid.in_(patient.pid.tolist()),
                Patient.death_time > start_time,
                Patient.death_time < end_time,
            )
        )
    )

    # merge tests with ukrdcids
    death_data = pd.read_sql(query_cod, session.bind)
    test_data_id_merged = death_data.merge(patient, "inner", on="pid")

    # turn into histogram
    test_data_hist = (
        test_data_id_merged[["ukrdcid", "cause_of_death"]]
        .drop_duplicates()
        .groupby(["cause_of_death"])
        .count()
        .reset_index()
    )

    return Labelled2d(
        metadata=Labelled2dMetadata(
            title="Cause of Death", coding_standard_x="EDTA_COD"
        ),
        data=Labelled2dData(
            x=list(test_data_hist.cause_of_death), y=list(test_data_hist.ukrdcid)
        ),
    )



"""
Patient cohort dialysis stats calculator
"""

import datetime as dt
from typing import Literal, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import BaseModel
from sqlalchemy import and_, func, or_, select
from sqlalchemy.orm import Session
from ukrdc_sqla.ukrdc import (
    DialysisSession,
    Patient,
    PatientRecord,
    Treatment,
)

from ukrdc_stats.abc import AbstractFacilityStatsCalculator
from ukrdc_stats.exceptions import NoCohortError

from .models.generic_2d import (
    AxisLabels2d,
    Labelled2d,
    Labelled2dData,
    Labelled2dMetadata,
)

from .models.networks import LabelledNetwork, NetworkMetaData, Nodes, Vertices


class DialysisStats(BaseModel):
    all_patients_home_therapies: LabelledNetwork
    incident_home_therapies: LabelledNetwork
    prevalent_home_therapies: LabelledNetwork
    incentre_dialysis_frequency: Labelled2d
    incident_initial_access: Labelled2d


def _calculate_frequency(
    from_time: dt.datetime, to_time: dt.datetime, no_of_events: int
):
    """calculates the frequency in per week units of events in a given timewindow

    Args:
        from_time (dt.datetime): start of window
        to_time (dt.datetime): end of window
        no_of_proceedures (int): no of things/events/proceedures which have occured

    Returns:
        _type_: frequency of events
    """
    delta_t = (to_time - from_time).days

    if delta_t > 0.0:
        return 7.0 * no_of_events / delta_t
    # else:
    # TODO: add proper error handling to this
    #    print("Time window is not positive and non-zero")

    return None


class DialysisStatsCalculator(AbstractFacilityStatsCalculator):
    """class to calcuate metrics associated with dialysis modalities"""

    def __init__(
        self,
        session: Session,
        facility: str,
        from_time: dt.datetime,
        to_time: dt.datetime,
    ):
        super().__init__(session, facility)

        # Create a precisely 2 element time window tuple
        self.time_window: Tuple[dt.datetime, dt.datetime] = (from_time, to_time)

    def _extract_base_patient_cohort(self) -> pd.DataFrame:
        """Extract a base patient cohort dataframe from the database

        Returns:
            pd.DataFrame: Patient cohort dataframe
        """

        patient_query = (
            select(
                PatientRecord.ukrdcid,
                Patient.pid,
                Treatment.admit_reason_code,
                Treatment.qbl05,
                Treatment.hdp04,
                Treatment.from_time,
                Treatment.to_time,
                Patient.death_time,
                Patient.dead,
            )  # type:ignore
            .join(Treatment, Treatment.pid == Patient.pid)  # type:ignore
            .join(PatientRecord, PatientRecord.pid == Patient.pid)  # type:ignore
            .where(
                and_(
                    # filter for facility
                    Treatment.health_care_facility_code == self.facility,
                    PatientRecord.sendingextract == "UKRDC",
                    # ensure patient is alive at beginning of time window
                    or_(
                        Patient.dead.is_(None), Patient.death_time > self.time_window[0]
                    ),
                    # filter on dialysis modalities
                    or_(
                        Treatment.admit_reason_code == "1",
                        Treatment.admit_reason_code == "2",
                        Treatment.admit_reason_code == "3",
                        Treatment.admit_reason_code == "5",
                        Treatment.admit_reason_code == "11",
                        Treatment.admit_reason_code == "12",
                    ),
                )
            )
        )

        return pd.read_sql(patient_query, self.session.bind)

    def _extract_incident_prevalent(self, base_cohort: pd.DataFrame) -> pd.DataFrame:
        """
        Takes a base cohort from _extract_base_patient_cohort and extracts the incident and prevalent patients.

        This is currently a draft version and probably needs careful reviewing.

        Args:
            base_cohort (pd.DataFrame): Base cohort from output of _extract_base_patient_cohort

        Returns:
            pd.DataFrame: Patient cohort dataframe
        """

        # If patients are alive and have not been discharged count them as prevalent
        base_cohort["prevalent"] = (
            pd.isnull(base_cohort.deathtime)
            | (base_cohort.deathtime > self.time_window[1])
        ) & ((base_cohort.totime > self.time_window[1]) | pd.isnull(base_cohort.totime))
        base_cohort.prevalent.fillna(False)

        # Get a list of patients to check for incidence status. All incident patients start within the timewindow.
        incident_ids = base_cohort[["ukrdcid"]][
            base_cohort.fromtime > self.time_window[0]
        ].drop_duplicates()

        # Run query to test if they have appeared as hd, pd, or Tx prior to beginning of window these will be discounted
        not_incident_ids_query = (
            select(PatientRecord.ukrdcid)
            .join(Treatment)
            .where(
                and_(
                    Treatment.admit_reason_code.in_(
                        ["1", "2", "3", "5", "11", "12", "20", "29", "78"]
                    ),
                    Treatment.from_time < self.time_window[0],
                    PatientRecord.ukrdcid.in_(incident_ids.ukrdcid.to_numpy()),
                )
            )
        )
        not_incident_ids = self.session.execute(not_incident_ids_query).all()

        # label patients identified in incident_ids who do not appear in previous group as incident
        incident_ids["incident"] = ~incident_ids.ukrdcid.isin(
            [id[0] for id in not_incident_ids]
        )

        # merge into patient cohort and replace NaN with false
        merged = pd.merge(base_cohort, incident_ids, how="left", on="ukrdcid")
        merged.incident = merged.incident.fillna(False)

        return merged

    def _calculate_therapy_types(
        self, scope: Literal["all", "incident", "prevalent"]
    ) -> Tuple[Nodes, Vertices]:
        """
        Breakdown of dialysis patients on home and in-centre therapies.
        The information is returned using pydantic classes designed handle
        networks (this is essentially what a sankey plot is)

        Args:
            Scope: allows stats to be calculated for incident, prevalent or all patients

        Returns:
            Nodes, Vertices: pydantic classes containing calculated data
        """
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # Filter patient cohort based on incident, prevalent or all
        patient_cohort: Union[pd.DataFrame, pd.Series]
        if scope == "all":
            patient_cohort = self._patient_cohort
        elif scope == "incident":
            patient_cohort = self._patient_cohort[self._patient_cohort.incident]
        elif scope == "prevalent":
            patient_cohort = self._patient_cohort[self._patient_cohort.prevalent]
        else:
            raise ValueError("Invalid scope")

        # filter patients in cohort with Haemodialysis modalities who are receiving therapies in hospital
        hosp_hd = len(
            patient_cohort[
                patient_cohort.admitreasoncode.isin(["1", "2", "3", "5"])
                & (patient_cohort.qbl05 == "HOSP")
            ]["ukrdcid"].drop_duplicates()
        )

        # filter "" on home therapies
        home_hd = len(
            patient_cohort[
                patient_cohort.admitreasoncode.isin(["1", "2", "3", "5"])
                & (patient_cohort.qbl05 == "HOME")
            ]["ukrdcid"].drop_duplicates()
        )

        # filter "" where database does not provide information
        na_hd = len(
            patient_cohort[
                patient_cohort.admitreasoncode.isin(["1", "2", "3", "5"])
                & patient_cohort.qbl05.isnull()
            ]["ukrdcid"].drop_duplicates()
        )

        # patients on peritoneal dialysis which presumably all happens at home
        home_pd = len(
            patient_cohort[patient_cohort.admitreasoncode.isin(["11", "12"])][
                "ukrdcid"
            ].drop_duplicates()
        )

        # assemble calculated numbers into the pydantic data structures used by the api
        nodes = Nodes(
            node_labels=[
                "Haemodialysis",
                "Peritoneal dialysis",
                "Home therapies",
                "In-centre therapies",
                "Incomplete/Not given",
            ]
        )

        vertices = Vertices(
            source=["0", "0", "0", "1"],
            target=["2", "3", "4", "2"],
            value=[str(home_hd), str(hosp_hd), str(na_hd), str(home_pd)],
        )

        return nodes, vertices

    def _calculate_dialysis_frequency(self):
        """
        Calculate the frequency with which dialysis occurs
        """
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # get list of hd patients
        patient_list = self._patient_cohort[
            (self._patient_cohort.admitreasoncode.isin(["1", "2", "3", "5"]))
            & (self._patient_cohort.qbl05 == "HOSP")
        ].ukrdcid.drop_duplicates()

        # get number of dialysis sessions per patient and the date of the first and last one
        query = (
            select(
                PatientRecord.ukrdcid,
                func.min(DialysisSession.procedure_time).label("fromtime"),
                func.max(DialysisSession.procedure_time).label("totime"),
                func.count(DialysisSession.procedure_type_code).label("sessioncount"),
            )
            .join(DialysisSession, DialysisSession.pid == PatientRecord.pid)
            .where(
                and_(
                    PatientRecord.ukrdcid.in_(patient_list),
                    DialysisSession.procedure_type_code == "302497006",  # filter for hd
                    DialysisSession.procedure_time > self.time_window[0],
                    DialysisSession.procedure_time < self.time_window[1],
                )
            )
            .group_by(PatientRecord.ukrdcid)
        )

        session_data = pd.read_sql(query, self.session.bind)

        # calculate frequency of dialysis
        session_data["freq"] = session_data.apply(
            lambda row: _calculate_frequency(
                row["fromtime"], row["totime"], row["sessioncount"]
            ),
            axis=1,
        )

        # turn into  histogram
        nbins = 15
        bins = np.linspace(0, 7, nbins)
        labels = [f"{bins[i-1]}- {bins[i]}" for i in range(1, nbins)]
        hist = pd.cut(session_data.freq, bins=bins, labels=labels).value_counts(
            sort=False
        )

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="In-Centre Dialysis Frequency",
                axis_titles=AxisLabels2d(
                    x="Frequency (days per week)", y="No. of Patients"
                ),
            ),
            data=Labelled2dData(
                x=list(hist.keys()), y=[int(value) for value in hist.values]
            ),
        )

    def _calculate_access_incident(self):
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        window = (
            select(
                PatientRecord.ukrdcid,
                DialysisSession.procedure_time,
                DialysisSession.qhd20,
                func.rank()
                .over(
                    order_by=DialysisSession.procedure_time,
                    partition_by=PatientRecord.ukrdcid,
                )
                .label("rnk"),
            )
            .join(DialysisSession, DialysisSession.pid == PatientRecord.pid)
            .where(
                PatientRecord.ukrdcid.in_(
                    # pylint: disable=singleton-comparison
                    self._patient_cohort[self._patient_cohort.incident == True].ukrdcid
                )
            )
        ).subquery()

        initial_access_query = (
            select(window.c.qhd20, func.count(window.c.ukrdcid).label("no"))
            .group_by(window.c.qhd20)
            .where(window.c.rnk == 1)
        )

        initial_access_data = pd.read_sql(initial_access_query, self.session.bind)

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Initial Vascular Access of Incident Patients",
                axis_titles=AxisLabels2d(x="Line Type", y="No. of Patients"),
            ),
            data=Labelled2dData(
                x=list(initial_access_data.qhd20), y=list(initial_access_data.no)
            ),
        )

    def _calculate_all_home_therapies(self):
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        all_patients_nodes, all_patients_vertices = self._calculate_therapy_types("all")

        return LabelledNetwork(
            metadata=NetworkMetaData(
                title="Proportion of all Dialysis Patients on Home Therapies"
            ),
            node=all_patients_nodes,
            link=all_patients_vertices,
        )

    def _calculate_incident_home_therapies(self):
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        incident_nodes, incident_vertices = self._calculate_therapy_types("incident")

        return LabelledNetwork(
            metadata=NetworkMetaData(
                title="Proportion of Incident Patients on Home Therapies"
            ),
            node=incident_nodes,
            link=incident_vertices,
        )

    def _calculate_prevalent_home_therapies(self):
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        prevalent_nodes, prevalent_vertices = self._calculate_therapy_types("prevalent")

        return LabelledNetwork(
            metadata=NetworkMetaData(
                title="Proportion of Prevalent Patients on Home Therapies"
            ),
            node=prevalent_nodes,
            link=prevalent_vertices,
        )

    def extract_patient_cohort(self):
        """
        Extract a complete patient cohort dataframe to be used in stats calculations
        """
        self._patient_cohort = self._extract_incident_prevalent(
            self._extract_base_patient_cohort()
        )

    def extract_stats(self) -> DialysisStats:
        """Extract all stats for the dialysis module

        Returns:
            DialysisStats: Dialysis statistics object
        """
        # If we don't already have a patient cohort, extract one

        if self._patient_cohort is None:
            self.extract_patient_cohort()

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # TODO: Do we want metadata like population size here?
        #       See DemographicsCalculator.extract_stats
        return DialysisStats(
            all_patients_home_therapies=self._calculate_all_home_therapies(),
            incident_home_therapies=self._calculate_incident_home_therapies(),
            prevalent_home_therapies=self._calculate_prevalent_home_therapies(),
            incentre_dialysis_frequency=self._calculate_dialysis_frequency(),
            incident_initial_access=self._calculate_access_incident(),
        )



"""
Abstract base classes for the ukrdc_stats package
"""

from abc import ABC, abstractmethod
from typing import Optional
import pandas as pd
from pydantic import BaseModel
from sqlalchemy.orm import Session


class AbstractFacilityStatsCalculator(ABC):
    """
    Abstract base class for facility stats calculators.

    We only enforce a couple of minor requirements:
    - The class must have a constructor that takes a database session and a facility code
    - The class must have an `extract_patient_cohort` method that assigns a pandas dataframe to the `_patient_cohort` attribute
    - The class must have a `calculate_stats` method that returns calculated stats as a pydantic model
    """

    def __init__(self, session: Session, facility: str):
        # Set up the database session
        self.session: Session = session
        # Store the facility code
        self.facility: str = facility
        # Create a pandas dataframe to store the results
        self._patient_cohort: Optional[pd.DataFrame] = None

    @abstractmethod
    def extract_patient_cohort(self) -> None:
        """
        Extract the patient cohort from the database, and assign it to self._patient_cohort
        """

    @abstractmethod
    def extract_stats(self) -> BaseModel:
        """
        Extract all stats from the patient cohort and return them in a Pydantic object

        Returns:
            BaseModel: Pydantic object containing all related stats
        """



"""
Patient cohort demographics stats calculator
"""

import datetime as dt
from typing import Optional, Dict
import pandas as pd

from sqlalchemy.orm import Session
from sqlalchemy import select, and_
from ukrdc_sqla.ukrdc import Patient, PatientRecord

from pydantic import BaseModel
from ukrdc_stats.abc import AbstractFacilityStatsCalculator
from ukrdc_stats.exceptions import NoCohortError

from ukrdc_stats.utils import age_from_dob

from ukrdc_stats.code_groupings import ETHNIC_GROUP_MAP, GENDER_GROUP_MAP


from .models.generic_2d import (
    Labelled2d,
    Labelled2dData,
    Labelled2dMetadata,
    AxisLabels2d,
)


class DemographicsMetadata(BaseModel):
    population: Optional[int] = None


class DemographicsStats(BaseModel):
    gender: Labelled2d
    ethnic_group: Labelled2d
    age: Labelled2d
    metadata: DemographicsMetadata


def _calculate_base_patient_histogram(
    cohort: pd.DataFrame, group: str, code_map: Optional[Dict] = None
) -> pd.DataFrame:
    """Extract a histogram of the patient cohort, grouped by the given column

    Args:
        cohort (pd.DataFrame): Patient cohort
        group (str): Column to group by

    Raises:
        NoCohortError: If the patient cohort is empty

    Returns:
        pd.DataFrame: Histogram dataframe of the patient cohort
    """

    if code_map:
        mapped_column = f"{group}_mapped"
        cohort[mapped_column] = cohort[group].map(code_map)

        histogram = (
            cohort[["ukrdcid", mapped_column]]
            .drop_duplicates()
            .groupby([mapped_column])
            .count()
            .reset_index()
        )

    else:
        histogram = (
            cohort[["ukrdcid", group]]
            .drop_duplicates()
            .groupby([group])
            .count()
            .reset_index()
        )

    return histogram.rename(columns={"ukrdcid": "Count"})


class DemographicsCalculator(AbstractFacilityStatsCalculator):
    """Calculates the demographics information based on the personal infomation listed in the patient table"""

    def __init__(
        self, session: Session, facility: str, date: Optional[dt.datetime] = None
    ):
        """Initialises the PatientDemographicStats class and immediately runs the relevant query

        Args:
            session (SQLAlchemy session): Connection to database to calculate statistic from.
            facility (str): Facility to calculate the
            date (datetime, optional): Date to calculate at. Defaults to today.
        """
        super().__init__(session, facility)

        # Set the date to calculate at, defaulting to today
        self.date: dt.datetime = date or dt.datetime.today()

    def _extract_base_patient_cohort(self) -> pd.DataFrame:
        """
        Extracts the patient cohort from the database into a pandas dataframe
        """

        # TODO: Add ability to filter on modality

        # select all patients with modalities that haven't finished
        patient_query = (
            select(
                PatientRecord.ukrdcid,
                Patient.gender,
                Patient.ethnic_group_code,
                Patient.birth_time,
            )  # type:ignore
            .join(PatientRecord, Patient.pid == PatientRecord.pid)  # type:ignore
            .where(
                and_(
                    PatientRecord.sendingextract == "UKRDC",
                    PatientRecord.sendingfacility == self.facility,
                )
            )
        )

        return pd.read_sql(patient_query, self.session.bind)

    def _calculate_gender(self) -> Labelled2d:
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        gender = _calculate_base_patient_histogram(
            self._patient_cohort, "gender", GENDER_GROUP_MAP
        )

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Gender Distribution",
                coding_standard_x="NHS_DATA_DICTIONARY",
                axis_titles=AxisLabels2d(x="Gender", y="No. of Patients"),
            ),
            data=Labelled2dData(
                x=gender.gender_mapped.tolist(), y=gender.Count.tolist()
            ),
        )

    def _calculate_ethnic_group_code(self):
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        ethnic_group_code = _calculate_base_patient_histogram(
            self._patient_cohort, "ethnicgroupcode", ETHNIC_GROUP_MAP
        )

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Ethnic Group",
                coding_standard_x="NHS_DATA_DICTIONARY",
                axis_titles=AxisLabels2d(x="Ethnicity", y="No. of Patients"),
            ),
            data=Labelled2dData(
                x=ethnic_group_code.ethnicgroupcode_mapped.tolist(),
                y=ethnic_group_code.Count.tolist(),
            ),
        )

    def _calculate_age(self):
        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # add column with ages and calculate histogram
        self._patient_cohort["age"] = self._patient_cohort["birthtime"].apply(
            lambda dob: age_from_dob(self.date, dob)
        )

        age = _calculate_base_patient_histogram(self._patient_cohort, "age")

        return Labelled2d(
            metadata=Labelled2dMetadata(
                title="Age Distribution",
                axis_titles=AxisLabels2d(x="Age", y="No. of Patients"),
            ),
            data=Labelled2dData(x=age.age.tolist(), y=age.Count.tolist()),
        )

    def extract_patient_cohort(self):
        """
        Extract a complete patient cohort dataframe to be used in stats calculations
        """
        self._patient_cohort = self._extract_base_patient_cohort()

    def extract_stats(self) -> DemographicsStats:
        """Extract all stats for the demographics module

        Returns:
            DemographicsStats: Demographics statistics object
        """
        # If we don't already have a patient cohort, extract one
        if self._patient_cohort is None:
            self.extract_patient_cohort()

        if self._patient_cohort is None:
            raise NoCohortError("No patient cohort has been extracted")

        # Crunch the numbers and make dataframes to produce "histograms" to display idividual bits of data
        pop_size = len(self._patient_cohort[["ukrdcid"]].drop_duplicates())

        # Build output object
        return DemographicsStats(
            metadata=DemographicsMetadata(population=pop_size),
            ethnic_group=self._calculate_ethnic_group_code(),
            gender=self._calculate_gender(),
            age=self._calculate_age(),
        )



"""
Functions to calculate patient cohort statistics.
Pydantic classes to return nice plotly friendly output.
Utility functions to handle bitty things.
"""

from . import demographics, models, utils

__all__ = ["demographics", "utils", "models"]



"""
Common utility functions useful in multiple statistics
"""

import datetime as dt


def age_from_dob(date: dt.date, dob: dt.date) -> int:
    """Returns the age on a given date

    Args:
        date (datetime): Date to calculate age or time period from.
        dob (datetime): Date to calculate age or time period at.

    Returns:
        int: age or period in years
    """
    years_old: int

    # calculates age by common definition
    years_old = date.year - dob.year - 1
    try:
        year_birthday = dt.datetime(date.year, dob.month, dob.day)
    except ValueError:
        # exemption triggered for people with birthday on leap year if not a leap year
        year_birthday = dt.datetime(date.year, dob.month, dob.day - 1)

    if year_birthday <= date:
        years_old += 1

    return years_old


def age_from_dob_exact(date: dt.date, dob: dt.date) -> float:
    """Generates an exact dob as decimal

    Args:
        date (dt.date): Date to calculate age or time period from.
        dob (dt.date): Date to calculate age or time period at.

    Returns:
        float: age
    """

    return (date - dob).days / 365.25


def dob_cutoff_from_age(date: dt.datetime, age: int) -> dt.datetime:
    """returns a date a fixed number of years before give date

    Args:
        date (dt.date): date to calculate from
        age (int): number of years before date

    Returns:
        dt.date: date a set number of years ago
    """

    return date - dt.timedelta(days=age * 365.25)



"""This file contains groupings and mappings of codes used by stats.
Maybe these should be in the database?
"""


# 5 ethnicity group mapping
ETHNIC_GROUP_MAP = {
    "A": "White",
    "B": "White",
    "C": "White",
    "D": "Mixed",
    "E": "Mixed",
    "F": "Mixed",
    "G": "Mixed",
    "H": "Asian",
    "J": "Asian",
    "K": "Asian",
    "L": "Asian",
    "M": "Black",
    "N": "Black",
    "P": "Black",
    "R": "Asian",
    "S": "Other",
    "Z": "Not Stated",
}

# NHS digital gender map
GENDER_GROUP_MAP = {"1": "Male", "2": "Female", "9": "Indeterminate", "X": "Unknown"}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ukrdc_stats', 'ukrdc_stats.models']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.40,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'psycopg2-binary>=2.9.3,<3.0.0',
 'pydantic>=1.10.1,<2.0.0',
 'ukrdc-sqla>=1.11.0,<2.0.0']

setup_kwargs = {
    'name': 'ukrdc-stats',
    'version': '0.1.0',
    'description': 'A package to produce stats from the ukrdc database to be displayed on the dashboard',
    'long_description': '# Dashboard Statistics Library\n\nLibrary for generating statistics for the ukrdc dashboard\n\n[![Test](https://github.com/renalreg/dashboard-stats/actions/workflows/main.yml/badge.svg)](https://github.com/renalreg/dashboard-stats/actions/workflows/main.yml)\n[![codecov](https://codecov.io/gh/renalreg/dashboard-stats/branch/master/graph/badge.svg?token=Ay8mk0zrKj)](https://codecov.io/gh/renalreg/dashboard-stats)\n\n## Developer notes\n\n### Installation\n\n```bash\npoetry install\n```\n\n### Iterating version numbers\n\nThe library should follow [semantic versioning](https://semver.org/).\n\n[Use Poetry to set the application version.](https://python-poetry.org/docs/cli/#version)\n\nE.g. `poetry version patch` for fix releases, `poetry version minor` for new functionality releases, or `poetry version major` for breaking-change releases.\n\n### Running the demo notebooks\n\nInstall additional demo notebook dependencies with\n\n```bash\npoetry install --with demo\n```\n',
    'author': 'Philip Main',
    'author_email': 'Philip.Main@renalregistry.nhs.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/renalreg/dashboard-stats',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Generic 2D data model for UKRDC stats
"""

from datetime import datetime
from typing import List, Optional, Union

from pydantic import BaseModel

Number = Union[int, float, None]

# Generics


class AxisLabels2d(BaseModel):
    """
    Generic class for any x/y axis labels
    """

    x: Optional[str] = None
    y: Optional[str] = None


class Basic2dMetadata(BaseModel):
    """
    Stats/plot metadata generic to all 2D data sets
    """

    title: Optional[str] = None
    axis_titles: Optional[AxisLabels2d] = None


# Time series


class TimeSeries2dData(BaseModel):
    """
    x-y data for a time series plot
    """

    x: List[datetime]
    y: List[Number]
    error_y: Optional[List[Number]] = None


class TimeSeries2dMetadata(Basic2dMetadata):
    """
    Metadata for a time series plot.

    x-axis is always a datetime, so no x-axis units are required.
    """

    units_y: Optional[str] = None


class TimeSeries2d(BaseModel):
    """
    Return data class for a time series plot
    """

    metadata: TimeSeries2dMetadata
    data: TimeSeries2dData


# Numeric


class Numeric2dData(BaseModel):
    """
    x-y data for a numeric plot
    """

    x: List[Number]
    y: List[Number]
    error_x: Optional[List[Number]] = None
    error_y: Optional[List[Number]] = None


class Numeric2dMetadata(Basic2dMetadata):
    """
    Metadata for a numeric plot.

    We may want units for both x and y values, so both are optional here.
    """

    units_x: Optional[str] = None
    units_y: Optional[str] = None


class Numeric2d(BaseModel):
    """
    Return data class for a numeric plot
    """

    metadata: Numeric2dMetadata
    data: Numeric2dData


class Labelled2dData(BaseModel):
    """
    x-y data for a labelled plot
    """

    x: List[str]
    y: List[Number]
    error_y: Optional[List[Number]] = None


class Labelled2dMetadata(Basic2dMetadata):
    """
    Metadata for a labelled plot.

    x-axis is always a string/label, so no x-axis units are required.
    y-axis is numeric, and so units are optional here.
    """

    coding_standard_x: Optional[str] = None
    units_y: Optional[str] = None


class Labelled2d(BaseModel):
    """
    Return data class for a labelled plot
    """

    metadata: Labelled2dMetadata
    data: Labelled2dData
